!=========================================================================
! Copyright (C) VMware, Inc. 1986-2011.  All Rights Reserved.
!
! $Id: ExceptionA.gs 26311 2011-08-01 22:42:15Z stever $
!
!=========================================================================

removeallmethods ExceptionA
removeallclassmethods ExceptionA

category: 'For Documentation Installation only'
classmethod: ExceptionA
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'Because GemStone/S already had an Exception class, the ANSI Exception 
class is named ExceptionA. This is not much of an inconvenience since 
most references will be to subclasses (Error, Warning, and Notification).

The class protocol describe the behavior of class objects that are used to 
create, signal, and select exceptions that exist within a specialization 
hierarchy.

The value of the standard global Exception is a class object that conforms 
to this protocol. The class Exception is explicitly specified to be 
subclassable. Conforming implementations must implement its behaviors in 
a non-fragile manner.

The instance protocol describes the behavior of instances of class Exception. 
Typically, actual exceptions used by an application will be either 
direct or indirect subclasses of this class. Exception combines the 
behavior of <exceptionBuilder> and <signaledException>. Instances are 
used to both supplied inform before an exception is signaled and to 
pass the information to an exception handler.

As Exception is explicitly specified to be subclassable, conforming 
implementations must implement its behavior in a non- fragile manner.'.
doc documentClassWith: txt.

self description: doc.
%

! ------------------- Class methods for ExceptionA
category: 'ANSI - exceptionSelector'
classmethod: ExceptionA
, anotherException
	"Returns an exception set that contains the receiver and the argument
	exception. This is commonly used to specify a set of exception selectors
	for an 'exception handler.'"

	^anotherException exceptionSetWithException: self.
%
category: 'ANSI support'
classmethod: ExceptionA
asSymbol

	^self name asSymbol.
%
category: 'ANSI support'
classmethod: ExceptionA
category: cat number: errNum arguments: args

	(cat isKindOf: ExceptionA) ifTrue: [^cat].
	cat == GemStoneError ifTrue: [
		^self 
			gsNumber: 		errNum 
			arguments: 	args.
	].
	^self new
		gsCategory:	cat;
		gsNumber:		errNum;
		gsArguments:	args;
		yourself.
%
category: 'ANSI support'
classmethod: ExceptionA
errorCategoryToCatch

	^self == ExceptionA ifTrue: [nil] ifFalse: [self]
%
category: 'ANSI support'
classmethod: ExceptionA
errorNumberToCatch

	^nil.
%
category: 'ANSI support'
classmethod: ExceptionA
errorNumberToSignal

	^1.
%
category: 'ANSI support'
classmethod: ExceptionA
exceptionSetWithException: anException

	^ExceptionSet
		with: anException
		with: self.
%
category: 'ANSI support'
classmethod: ExceptionA
exceptionSetWithExceptionSet: anExceptionSet

	anExceptionSet add: self.
	^anExceptionSet.
%
category: 'ANSI support'
classmethod: ExceptionA
gsNumber: errNum arguments: args

	(errNum >= 2001 and: [errNum <= 2999]) ifTrue: [
	    errNum == 2010 ifTrue: [
			^MessageNotUnderstood new
				message: (FailedMessage new
					arguments: (args at: 3);
					selector: (args at: 2);
					yourself);
				receiver: (args at: 1);
				gsCategory: GemStoneError;
				gsNumber: errNum;
				gsArguments: args;
				yourself.
			].
			errNum == 2026 ifTrue: [
				^ZeroDivide new
					tag: args first;
					gsCategory: GemStoneError;
					gsNumber: errNum;
					gsArguments: args;
					yourself.
			].
		].
	^self new
		gsCategory: GemStoneError;
		gsNumber: errNum;
		gsArguments: args;
		yourself.
%
category: 'ANSI - exceptionSelector'
classmethod: ExceptionA
handles: anExceptionSelector

	"This message determines whether the 'exception handler' associted with
	the receiver may be used to process the argument. Answer 'true' if an 
	associated handler should be used to process 'exception.' Answer 'false'
	if an associated handler may not be used to process the exception.

	Returns 'true' if the class of 'exception' is the receiver or a general
	subclass of the receiver.

	This definition implies that subclasses of an exception class are considered 
	to be 'subexceptions' of the type of exception defined by their superclass.
	An 'exception handler' that 'handles' an exception class will also handle any
	exceptions that are instances of the exception class's subclasses."

	^self == ExceptionA or: [anExceptionSelector isKindOf: self]
%
category: 'ANSI support'
classmethod: ExceptionA
handlesCategory: anObject number: anInteger
	"This method may be overridden for ExceptionSelectors that are more discriminating.
	For example, one might want to trap all GemStoneError numbers in a certain range."

	^self handles: anObject.
%
category: 'ANSI support'
classmethod: ExceptionA
isGemStoneError

	^#[
		Exception ,
		ExceptionA ,
		Error , 
		MessageNotUnderstood ,
		ZeroDivide 
	] includes: self.
%
category: 'ANSI - exceptionInstantiator'
classmethod: ExceptionA
new
	"Return a newly created object initialized to a standard initial state.

	The object returned is an <exceptionBuilder> that may be used to
	signal an exception of the same type that would be signaled if the
	message #signal is sent to the receiver.

	The object returned conforms to <Exception>."

	^self basicNew initialize
%
category: 'ANSI - exceptionInstantiator'
classmethod: ExceptionA
signal

	"An exception of the type associated with the receiver is signaled.
	The <signaledException> is initialized to its default state.

	The exception signaled conforms to <Exception> with all of its
	<exceptionDescription> attributes set to their default values."

	^self signal: nil.
%
category: 'ANSI - exceptionSignaler'
classmethod: ExceptionA
signal: signalText
	
	^self new signal: signalText.
%
category: 'bonus'
classmethod: ExceptionA
signal: signalText with: tagObject
	
	^self new signal: signalText with: tagObject.
%
category: 'bonus'
classmethod: ExceptionA
signalWith: tagObject
	
	^self new signal: nil with: tagObject.
%
category: 'SUnit'
classmethod: ExceptionA
sunitSignalWith: aString
	^self signal: aString
%
! ------------------- Instance methods for ExceptionA
category: 'ANSI - exceptionDescription'
method: ExceptionA
defaultAction
	"If the exception described by the receiver is signaled 
	and the current <exception environment> does not
	contain a handler for the exception, this method will
	be executed."

	| cat num args |
	cat 		:= self gsCategory.
	num 	:= self gsNumber.
	args 	:= self gsArguments.
	(cat isKindOf: SymbolDictionary) ifFalse: [
		cat := SymbolDictionary new
			at: #'English' put: #(#(1));
			yourself.
		num := 1.
		args := Array with: self genericDescription.
	].
	^System
		_signalGciError: 	num
		args: 						args
		signalDictionary: 	cat.
%
category: 'ANSI - exceptionDescription'
method: ExceptionA
description
	"Return text that describes in a human readable form 
	an occurrence of an exception. If an explicit message
	text was provided by the signaler of the exception,
	that text should be incorporated into the description."

	| cat array num desc args |
	messageText ~~ nil ifTrue: [^messageText].
	(cat := self gsCategory) isNil ifTrue: [^self genericDescription].
	(cat isKindOf: SymbolDictionary) ifFalse: [^self genericDescription].
	array := cat 
		at: System myUserProfile nativeLanguage 
		ifAbsent: [^self genericDescription].
	(num := self gsNumber) isNil ifTrue: [^self genericDescription].
	array size < num ifTrue: [^self genericDescription].
	(desc := array at: num) isNil ifTrue: [^self genericDescription].
	((args := self gsArguments) isKindOf: Array) ifFalse: [^self genericDescription].
	^self gsCategory 
		textForError: num
		args: args.
%
category: 'ANSI support'
method: ExceptionA
genericDescription

	^messageText notNil
		ifTrue: [messageText]
		ifFalse: [self class name].
%
category: 'Accessors'
method: ExceptionA
gsArguments

	^gsArguments.
%
category: 'Accessors'
method: ExceptionA
gsArguments: anArray

	gsArguments := anArray.
%
category: 'Accessors'
method: ExceptionA
gsCategory

	^gsCategory.
%
category: 'Accessors'
method: ExceptionA
gsCategory: anExceptionSelector

	gsCategory := anExceptionSelector.
%
category: 'Accessors'
method: ExceptionA
gsNumber

	^gsNumber ~~ nil
		ifTrue:  [gsNumber]
		ifFalse: [self class errorNumberToSignal].
%
category: 'Accessors'
method: ExceptionA
gsNumber: anInteger

	gsNumber := anInteger.
%
category: 'Accessors'
method: ExceptionA
handler

	^handler.
%
category: 'Accessors'
method: ExceptionA
handler: anExceptionHandler

	handler := anExceptionHandler.
%
category: 'ANSI support'
method: ExceptionA
initialize

	gsCategory 		:= self.
	gsNumber 		:= nil.
	gsArguments 	:= #().
%
category: 'ANSI - signaledException'
method: ExceptionA
isNested
	"Answer 'true' if the 'handler environment' for the current exception handler
	contains an 'exception handler' that will 'handle' the receiver. Answer 'false'
	if it does not.

	The 'default action' for an exception is not considered to be an enclosing handler.
	Only the existence of a handler explicitly established using #on:do: will result in 
	this method returning 'true'."

	^handler isNested.
%
category: 'ANSI - exceptionDescription'
method: ExceptionA
isResumable

	^false.
%
category: 'ANSI - exceptionDescription'
method: ExceptionA
messageText
	"Return the signaler message text of the receiver. If the signaler
	has not provided any message text, return nil."

	^messageText
%
category: 'ANSI - exceptionBuilder'
method: ExceptionA
messageText: signalerText
	"Set the signaler message text of the receiver. Subsequent sends of the 
	message #messageText to the receiver will return this value.
	Subsequent sends of the message #messageText to a 'signaled exception'
	generated by sending the message #signal to the receiver of this message
	will also return this value."


	messageText := signalerText.
%
category: 'Accessors'
method: ExceptionA
number

	^self gsNumber.
%
category: 'ANSI - signaledException'
method: ExceptionA
outer

	"If the 'handler environment' for the current exception handler contains an
	'exception handler' that will 'handle' the receiver, evaluate that handler's
	'exception action' with the receiver as the argument to its 'handler block.' 
	If there is no enclosing handler, send the message #defaultAction to the
	receiver. The #defaultAction method is evaluated using the current
	'exception environment.'

	If the receiver is 'resumable' and the evaluated 'exception action' resumes 
	then the result returned from #outer will be the 'resumption value' of the
	evaluated 'exception action.' If the receiver is not 'resumable' or if the
	'exception action' does not resume then this message will not return.

	For exceptions that are not 'resumable', #outer is equivalent to #pass."

	^handler outer.
%
category: 'ANSI - signaledException'
method: ExceptionA
pass
	"If the 'handler environment' for the current exception handler contains an
	enclosing 'exception handler' for the receiver, activate that handler's 
	'exception action' in place of the current 'exception action.' If there is no
	enclosing handler, execute the 'default action' for the receiver as if no
	handler had been found when the exception was originally signaled. The
	default action is evaluated in the context of the signaling environment.

	Control does not return to the currently active 'exception handler.'"

	handler resume: handler outer.
	self error: 'Should not get here!'.
%
category: 'ANSI - signaledException'
method: ExceptionA
resignalAs: replacementException
	"The active 'exception action' is aborted and the 'exception environment'
	and the 'evaluation context' are restored to the same states that were in
	effect when the receiver was originally signaled. Restoring the 'evaluation
	context' may result in the execution of #ensure: or #ifCurtailed: termination
	blocks.

	After restoration, signal the 'replacementException' and execute the
	'exception action' as determined by the restored 'exception environment.'

	This message causes the 'replacementException' to be treated as if it had
	been originally signaled instead of the receiver.

	If the 'replacementException' is 'resumable' and its 'exception action' resumes,
	control will ultimately return from the message that signaled the original 
	exception.

	Control does not return from this message to the currently active
	'exception action.'"

	handler resignalAs: replacementException.
	self error: 'Should not get here!'.
%
category: 'ANSI - signaledException'
method: ExceptionA
resume
	"If the current 'exception action' was activated as the result of sending
	the message #outer to the receiver, return a resumption value as the 
	value of the #outer message.

	If the receiver is a 'resumable' exception a resumption value is returned
	as the value of the message that signaled the receiver. Before returning,
	the 'exception environment' and the 'evaluation context' are restored
	to the same states that were in effect when the receiver was originally
	signaled. Restoring the 'evaluation context' may result in the execution of
	#ensure: or #ifCurtailed: termination blocks.

	This message does not return to its point of invocation.

	The resumption value is unspecified."

	self resume: nil.
%
category: 'ANSI - signaledException'
method: ExceptionA
resume: resumptionValue
	"If the current 'exception action' was activated as the result of sending the
	message #outer to the receiver, return 'resumptionValue' as the value
	of the #outer message.

	If the receiver is a 'resumable' exception, the 'resumptionValue' is returned
	as the value of the message that signaled the receiver. Before returning,
	the 'exception environment' and the 'evaluation context' are restored to the 
	same states that were in effect when the receiver was originally signaled.
	Restoring the 'evaluation context' may result in the execution of #ensure:
	or #ifCurtailed: termination blocks.

	This message does not return to its point of invocation."

	self isResumable ifFalse: [self error: 'Attempt to resume from a non-resumable exception'].
	handler resume: resumptionValue.
%
category: 'ANSI - signaledException'
method: ExceptionA
retry
	"The active 'exception action' is aborted and the 'exception environment' and the
	'evaluation context' are restored to the same states that were in effect when the 
	#on:do: message that established the active handler was sent. Restoring the
	'evaluation context' may result in the execution of #ensure: or #ifCurtailed: 
	termination blocks.

	After the restoration, the #on:do: method is re-evaluated with its original receiver
	and arguments. Control does not return from this message to the curently active
	'exception action.'"

	handler retryUsing: nil.
%
category: 'ANSI - signaledException'
method: ExceptionA
retryUsing: alternativeBlock
	"The active 'exception action' is aborted and the 'exception environment' and the
	'evaluation context' are restored to the same states that were in effect when the
	#on:do: message that established the active handler was sent. Restoring the 
	'evaluation context' may result in the execution of #ensure: or #ifCurtailed blocks.

	After the restoration, the #on:do: method is re-evaluated with 'alternativeBlock'
	substituted for its original receiver. The original arguments are used for the 
	re-evaluation.

	Control does not return from this message to the currently active 'exception action.'"

	alternativeBlock isNil ifTrue: [self error: 'Invalid argument to #retryUsing:'].
	handler retryUsing: alternativeBlock.
%
category: 'ANSI - signaledException'
method: ExceptionA
return
	"The value 'nil' is returned as the value of the 'protected block' of the active
	'exception handler.' Before returning, the 'exception environment' and the 
	'evaluation context' are restored to the same states that were in effect when
	the active handler was created using #on:do:. Restoring the 'evaluation context'
	may result in the execution of #ensure: or #ifCurtailed: termination blocks.

	This message does not return to its point of invocation."

	self return: nil
%
category: 'ANSI - signaledException'
method: ExceptionA
return: returnValue
	"The 'returnValue' is returned as the value of the 'protected block' of the 
	active 'exception handler.' Before returning, the 'exception environment'
	and the 'evaluation context' are restored to the same states that were in 
	effect when the active handler was created using #on:do:. Restoring the
	'evaluation context' may result in the execution of #ensure: or #ifCurtailed
	termination blocks.

	This message does not return to its point of invocation."

	handler return: returnValue.
%
category: 'ANSI - exceptionSignaler'
method: ExceptionA
signal
	"Associated with the receiver is an <exceptionDescription> called the 'signaled exception.'
	The current 'exception environment' is searched for an 'exception handler' whose
	'exception selector' mathces the 'signaled exception.' The search proceeds from the most
	recently created 'exception handler' to the oldest 'exception handler.'

	A matching handler is defined to be one which would return 'true' if the message #handles:
	was sent to its 'exception selector' with the 'signaled exception' as the argument.

	If a matching handler is found, the 'exception action' of the handler is evaluated in the 
	'exception environment' tat was current when the handler was created and the state of
	the curent 'exception environment' is preserved as the 'signaling environment.'

	The 'exception action' is evaluated as if the message #value: were sent to it with a
	<signaledException> passed as its argument. The <signaledException> is derived from
	the 'signaled exception' in an implementation dependent manner.

	If the evaluation of the 'exception action' returns normally (as if it had returned from the
	#value: message), the 'handler environment' is restored and the value returned from the
	'exception action' is returned as the value of the #on:do: message that created the
	handler. Before returning, any active #ensure: or #ifCurtailed: termination blocks created
	during evaluation of the receiver of the #on:do: message are evaluated.

	If a matching handler is not found when the 'exception environment' is searched, the
	'default action' for the 'signaled exception' is performed. This is accomplished as if the
	message #defaultAction were sent to the <signaledException> object derived from the 
	'signaled exception.' If the 'exception' is 'resumable' the value returned from the 
	#defaultAction method is returned as the value of the #signal message. If the 
	'signaled exception' is not 'resumable' the action taken upon completion of the
	#defaultAction method is implementation defined."

	| gsException |
	gsException := Exception 
		_exceptionFor: self gsNumber
		signalDictionary: self gsCategory.
	gsException isNil ifTrue: [
		^self _defaultAction.
	].
	handler := gsException handler.
	^gsException
		_signal: self gsCategory
		number: self gsNumber
		args: self gsArguments.
%
category: 'ANSI - exceptionSignaler'
method: ExceptionA
signal: signalText
	"Associated with the receiver is an <exceptionDescription> called the 'signaled exception.'
	The message text of the signaled exception is set to the value of signalerText, and then
	the exception is signaled in the same manner as if the message #signal had been sent
	to the receiver.

	Note that this message does not return in some circumstances. The situations in which
	it does not return and the returned value, if any, are the same as specified for the 
	#signal message."

	messageText := signalText.
	^self signal.
%
category: 'bonus'
method: ExceptionA
signal: signalText with: tagObject

	tag := tagObject.
	^self signal: signalText.
%
category: 'SUnit'
method: ExceptionA
sunitExitWith: aValue
	^self return: aValue
%
category: 'SUnit'
method: ExceptionA
sunitSignalWith: aString
	^self signal: aString
%
category: 'ANSI - exceptionDescription'
method: ExceptionA
tag
	"Return the tag value provided by the signaler of the receiver.
	If the signaler has not provided a tag value, return the same
	value as would be returned if #messageText was sent to the
	receiver of this message. If the signaler has provided neither
	a tag value nor a message text, return nil.

	Exception tags are intended for use in situations where a 
	particular occurrence of an exception needs to be identified
	and a textual description is not appropriate. For example,
	the message text might vary according to the locale and 
	thus could not be used to identify the exception."

	^tag ~~ nil 
		ifTrue: [tag]
		ifFalse: [self messageText].
%
category: 'Accessors'
method: ExceptionA
tag: anObject

	tag := anObject.
%
category: 'Accessors'
method: ExceptionA
textForError: anInteger args: anArray
	"For polymorphism with SymbolDictionary"
	
	^self genericDescription.
%
category: 'ANSI support'
method: ExceptionA
_defaultAction

	| myResult |
	myResult := self defaultAction.
	self isResumable ifTrue: [^myResult].

"If the <signaled exception> is not <resumable> the action taken 
upon completion of the #defaultAction method is implementation defined.
-- ANSI section 5.5.2.1"

	Error signal: 'Returned from the default action of a non-resumable exception'.
%

